#include <iostream>
#include <fstream>
#include <cstring>
#include <sstream>
#include <iomanip>
#include <cstdlib>

#include "pak.h"
#include "util.h"
using namespace std;

static const char* pngStartSig="\x89PNG";
static const char* pngEndSig="IEND\xae\x42\x60\x82";
char* outPath=0;
static bool verbose=0;
static bool debug=0;

int getResourceOffset(char* descriptorOffset,int version,int& id,int& size,char*& nextDescriptor)
{
    switch(version)
    {
        case 1:
        {
            if(debug)
                cerr << "extracting pak version 1\n";
            resourceV1HeaderEntry* entry((resourceV1HeaderEntry*)descriptorOffset);
            id=entry->id;
            size=entry->size;
            nextDescriptor=(char*)&entry[1];
            if(debug)
                cerr << "getResource: id="<<hex<<id<<", offset="<<entry->offset<<", size="<<size<<endl;
            return entry->offset;
        }
        case 4:
        {
            if(debug)
                cerr << "extracting pak version 4\n";
            resourceV4HeaderEntry* entry((resourceV4HeaderEntry*)descriptorOffset);
            id=entry->id;
            if(id)
                size=entry[1].offset-entry->offset;
            else
                size=0; // ID0 is last entry with zero size
            nextDescriptor=(char*)&entry[1];
            if(debug)
                cerr << "getResource: id="<<hex<<id<<", offset="<<entry->offset<<", size="<<size<<endl;
            return entry->offset;
        }
        default:
        {
            cerr << "Package format version " << version <<" is not supported!\n";
            exit(2);
        }
    }
}

void writeResourceToFile(int id, char* res, int size,int offset,const char* extension=".dat")
{
    ofstream file;
    ostringstream filename;
    filename <<outPath<<"/"<<hex<<setw(4)<<setfill('0')<<id<<extension;
    if(debug)
        cerr << "writing " << filename.str().c_str()<<endl;
    file.open(filename.str().c_str(),ios::out|ios::binary|ios::trunc);
    file.write(res,size);
    file.close();
}

bool extractPak(const char* filename)
{
    int pakSize;
    char* pakData=readFile(filename,pakSize);

    char* p(pakData);

    int version=*((int*)p);
    if(verbose)
        cerr << "pak version: " << version<<endl;
    p+=4;
    int numOfEntries=*((int*)p);
    if(verbose)
        cerr << "entries count: " << numOfEntries<<endl;
    p+=4;
    if(version==4)
    {
        cerr << "encoding: " << (unsigned int)*p<<endl;
        ++p; // skip encoding
    }

    int id=0,size=0;
    for(int i=0;i<numOfEntries;++i)
    {
        char* res=pakData+getResourceOffset(p,version,id,size,p);
        if(debug)
            cerr << "offset: "<<res-pakData<<", data: " << hex<<(int*)res<<endl;
        if(bstrstr(res,pngStartSig,size)==res)
        {
            if(verbose)
                cerr << "png found: id="<<hex<<id<<", size="<<size<<endl;
            char* foundEndSigOffset=bstrstr(res,pngEndSig,size);
            int pngSize=foundEndSigOffset+strlen(pngEndSig)-res;
            if(size!=pngSize)
            {
                cerr << "ERROR: png size doesn't match resource size!!! res size: "<<size<<"; png size: "<<pngSize<<endl;
                continue;
            }
            writeResourceToFile(id,res,size,res-pakData,".png");
        }
        else
        {
            if(verbose)
                cerr << "unknown resource type: id="<<hex<<id<<", size="<<size<<endl;
            writeResourceToFile(id,res,size,res-pakData);
        }
    }

    return true;
}

int main(int argc, char** argv)
{
    static const char* options="-v\tverbose\n-d\tdebug mode";
    if(argc<3)
    {
        cerr << "Usage:\n\t"<<argv[0]<<" [options] infile.pak /out/directory\nNote: /out/directory path must exist\nOptions:\n"<<options<<"\n";
        return 1;
    }
    int i;
    for(i=1;i<=argc;++i)
    {
        if(argv[i][0]=='-')
        {
            switch(argv[i][1])
            {
                case 'v':
                    verbose=true;
                    break;
                case 'd':
                    debug=true;
                    break;
                default:
                    cerr << "Unknown option: \""<<argv[i]<<"\". Valid options:\n"<<options<<endl;
                    return 1;
            }
        }
        else
            break;
    }
    outPath=argv[i+1];
    return !extractPak(argv[i]);
}
