#include <cstring>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>

#include "pak.h"
#include "util.h"
using namespace std;

char* outFile=0;
static bool verbose=false;
static bool debug=false;

static int version=4;

bool pack(const char* path)
{
    vector<string>& files(*getdir(path));

    switch(version)
    {
        case 1:
        {
            int headerSize=sizeof(headerV1)+files.size()*sizeof(resourceV1HeaderEntry);

            // Fill header version and numOfEntries
            char* header=new char[headerSize];
            headerV1& h(*(headerV1*)header);
            h.version=version;
            h.numOfEntries=files.size();
            resourceV1HeaderEntry* entries=(resourceV1HeaderEntry*)(header+sizeof(h));

            // Read data from all files
            vector<char*> dataPointers;
            resourceV1HeaderEntry* headerEntry(entries);
            int totalDataSize=0;

            for(vector<string>::const_iterator i=files.begin();i!=files.end();++i)
            {
                // The actual file data
                string filepath=string(path)+"/"+*i;
                if(debug)
                    cerr << "reading file " << filepath.c_str()<<"; ";
                char* data=readFile(filepath.c_str(),headerEntry->size);
                // Save pointer to data, this data will be merged into single file after we read all input files
                dataPointers.push_back(data);

                // Header entry for current file
                istringstream is(*i);
                is >> hex >> headerEntry->id; // This ignores filename extension, basename is used as hex number
                headerEntry->offset=headerSize+totalDataSize;
                if(debug)
                    cerr << "resource ID: " << hex << headerEntry->id<<"; size: "<<headerEntry->size<<"; offset: "<<headerEntry->offset << endl;
                totalDataSize+=headerEntry->size;

                ++headerEntry; // go to next entry in array
            }

            // Now merge the data
            char* outData=new char[totalDataSize];
            resourceV1HeaderEntry* currentEntry(entries);
            for(vector<char*>::const_iterator i=dataPointers.begin();i!=dataPointers.end();++i)
            {
                if(debug)
                    cerr << "memcpy(dest-outData+headerSize="<<currentEntry->offset<<", size="<<currentEntry->size<<")\n";
                memcpy(outData+currentEntry->offset-headerSize,*i,currentEntry->size);
                ++currentEntry;
            }

            // And finally write out the data
            ofstream file;
            file.open(outFile,ios::out|ios::binary|ios::trunc);
            file.write(header,headerSize);
            file.write(outData,totalDataSize);
            file.close();

            delete[] header;
            delete[] outData;
        }
        break;

        case 4:
        {
            int headerSize=sizeof(headerV4)+(1+files.size())*sizeof(resourceV4HeaderEntry);

            // Fill header version and numOfEntries
            char* header=new char[headerSize];
            headerV4& h(*(headerV4*)header);
            h.version=version;
            h.numOfEntries=files.size();
            h.encoding=1;
            resourceV4HeaderEntry* entries=(resourceV4HeaderEntry*)(header+sizeof(h));

            // Read data from all files
            vector<char*> dataPointers;
            resourceV4HeaderEntry* headerEntry(entries);
            int totalDataSize=0;

            for(vector<string>::const_iterator i=files.begin();i!=files.end();++i)
            {
                // The actual file data
                string filepath=string(path)+"/"+*i;
                if(debug)
                    cerr << "reading file " << filepath.c_str()<<"; ";
                int size;
                char* data=readFile(filepath.c_str(),size);
                if(debug)
                    cerr << dec<<size<<" bytes read\n";
                // Save pointer to data, this data will be merged into single file after we read all input files
                dataPointers.push_back(data);

                // Header entry for current file
                istringstream is(*i);
                is >> hex >> headerEntry->id; // This ignores filename extension, basename is used as hex number
                if(debug)
                    cerr << "resource ID: " << hex << headerEntry->id << endl;
                headerEntry->offset=headerSize+totalDataSize;

                ++headerEntry; // go to next entry in array

                totalDataSize+=size;
            }
            // Final id0 entry
            headerEntry->id=0;
            headerEntry->offset=totalDataSize+headerSize;

            // Now merge the data
            char* outData=new char[totalDataSize];
            resourceV4HeaderEntry* currentEntry(entries);
            for(vector<char*>::const_iterator i=dataPointers.begin();i!=dataPointers.end();++i)
            {
                int size=currentEntry[1].offset-currentEntry->offset;
                if(debug)
                    cerr << "memcpy(dest-outData="<<currentEntry->offset<<", size="<<dec<<size<<")\n";
                memcpy(outData+currentEntry->offset-headerSize,*i,size);
                ++currentEntry;
            }

            // And finally write out the data
            ofstream file;
            file.open(outFile,ios::out|ios::binary|ios::trunc);
            file.write(header,headerSize);
            file.write(outData,totalDataSize);
            file.close();

            delete[] header;
            delete[] outData;
        }
        break;

        default:
        {
            cerr << "Package format version " << version <<" is not supported!\n";
            exit(2);
        }
    }

    delete &files;
    return true;
}

int main(int argc, char** argv)
{
    static const char* options="-v\tverbose\n-d\tdebug mode\n-1\tuse pak version 1 instead of default 4";
    if(argc<3)
    {
        cerr << "Usage:\n\t"<<argv[0]<<" [options] /input/directory outfile.pak\nOptions:\n"<<options<<"\n";
        return 1;
    }
    int i;
    for(i=1;i<=argc;++i)
    {
        if(argv[i][0]=='-')
        {
            switch(argv[i][1])
            {
                case 'v':
                    verbose=true;
                    break;
                case 'd':
                    debug=true;
                    break;
                case '1':
                    version=1;
                    break;
                default:
                    cerr << "Unknown option: \""<<argv[i]<<"\". Valid options:\n"<<options<<endl;
                    return 1;
            }
        }
        else
            break;
    }
    outFile=argv[i+1];
    return !pack(argv[i]);
}
